import copy
from unittest.mock import Mock

import pytest

from tavern._core.pytest.config import TavernInternalConfig, TestConfig
from tavern._core.strict_util import StrictLevel

# GraphQL-specific configuration
_graphql_includes = TestConfig(
    variables={},
    strict=StrictLevel.all_on(),
    tavern_internal=TavernInternalConfig(
        pytest_hook_caller=Mock(),
        backends={"graphql": "gql"},
    ),
    follow_redirects=False,
    stages=[],
)


@pytest.fixture(scope="function", name="graphql_test_block_config")
def graphql_test_block_config_fixture():
    """Test block configuration for GraphQL tests."""
    return copy.deepcopy(_graphql_includes)



import asyncio
import json
import logging
from collections.abc import AsyncGenerator
from dataclasses import dataclass, field
from typing import Any, Optional, Union

from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.exceptions import TransportQueryError, TransportServerError
from gql.transport.websockets import WebsocketsTransport
from graphql import ExecutionResult

from tavern._core.asyncio import ThreadedAsyncLoop
from tavern._plugins.common.response import ResponseLike

logger: logging.Logger = logging.getLogger(__name__)

_SubResponse = (
    AsyncGenerator[dict[str, Any], None] | AsyncGenerator[ExecutionResult, None]
)
"""The type of a response from a subscription."""

ResultOrErr = Union[ExecutionResult, TransportQueryError]
"""The type returned from a gql query"""


class ClientCacheKey:
    """Cache key for GraphQL clients.

    Uses a tuple of sorted header items to ensure hashability.
    """

    url: str
    headers: tuple[tuple[str, str], ...]
    timeout: int

    def __init__(self, url: str, headers: dict[str, str], timeout: int):
        """Initialize ClientCacheKey with conversion of headers dict to tuple.

        Args:
            url: The URL for the transport
            headers: Headers dictionary
            timeout: Timeout in seconds
        """
        # Convert headers dict to sorted tuple for hashability
        self.url = url
        self.headers = tuple(sorted(headers.items()))
        self.timeout = timeout

    def __hash__(self) -> int:
        """Make ClientCacheKey hashable."""
        return hash((self.url, self.headers, self.timeout))

    def __eq__(self, other: object) -> bool:
        """Compare two ClientCacheKey instances."""
        if not isinstance(other, ClientCacheKey):
            return NotImplemented
        return (
            self.url == other.url
            and self.headers == other.headers
            and self.timeout == other.timeout
        )

    def __repr__(self) -> str:
        """String representation of ClientCacheKey."""
        return f"ClientCacheKey(url={self.url!r}, headers={self.headers!r}, timeout={self.timeout})"


@dataclass(kw_only=True)
class GraphQLResponseLike(ResponseLike):
    """A response-like object implementing the ResponseLike protocol for GraphQL responses"""

    result: ResultOrErr
    headers: dict[str, str] = field(default_factory=dict)

    @property
    def text(self) -> str:
        """Return the JSON serialized representation of the GraphQL result data.

        Returns:
            A JSON string representation of the result data.
        """
        if not self.result:
            raise ValueError("No GraphQL result to return")
        return json.dumps(self.result.data)

    def json(self) -> Any:
        """Parse and return the JSON content of the response"""
        return self.result.data


class GraphQLClient:
    """GraphQL client for HTTP requests and subscriptions over WebSocket"""

    _subscriptions: dict[str, _SubResponse]
    """Dictionary of active subscriptions, keyed by subscription name."""

    _to_close: list[AsyncGenerator]
    """List of subscription generators to close when the client is closed."""

    _gql_client_cache: dict[ClientCacheKey, Client]
    """Cache of GraphQL clients to avoid creating new ones for the same URL."""

    _transport_cache: dict[ClientCacheKey, AIOHTTPTransport]
    """Cache of HTTP transports associated with each client."""

    def __init__(self, **kwargs):
        """Initialize the GraphQL client."""
        self.default_headers = kwargs.get("headers", {})
        self.timeout = kwargs.get("timeout", 30)

        self._subscriptions = {}
        self._to_close = []
        self._gql_client_cache = {}
        self._transport_cache = {}

        self._threaded_async_loop = ThreadedAsyncLoop()

    def __enter__(self):
        """Enter the context manager.

        Returns:
            The GraphQLClient instance.
        """
        self._threaded_async_loop.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context manager and close WebSocket connections."""

        async def _close_subscriptions():
            """Close all active subscription generators."""
            await asyncio.gather(*(s.aclose() for s in self._subscriptions.values()))
            await asyncio.gather(*(s.aclose() for s in self._to_close))
            await asyncio.gather(*(s.close() for s in self._transport_cache.values()))

        try:
            # Schedule the closing of subscriptions in the event loop and
            # wait for the closing operations to complete with a timeout
            self._threaded_async_loop.run_coroutine(_close_subscriptions(), timeout=5.0)
        except TimeoutError:
            logger.warning("Timed out waiting for subscriptions to close")

        # Join the thread with a timeout to prevent indefinite blocking
        self._threaded_async_loop.__exit__(exc_type, exc_val, exc_tb)

    def make_request(
        self,
        url: str,
        query: str,
        variables: Optional[dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        headers: Optional[dict] = None,
        has_files: bool = False,
    ) -> ResponseLike:
        """Execute GraphQL query/mutation over HTTP using gql.

        Args:
            url: The GraphQL endpoint URL.
            query: The GraphQL query string.
            variables: Optional variables for the query.
            operation_name: Optional name of the operation to execute.
            headers: any headers to send with the request
            has_files: whether the request contains files

        Returns:
            A GraphQLResponseLike object containing the query result.

        Raises:
            Exception: If the request fails or returns errors.
        """
        headers = headers or {}
        headers = dict(self.default_headers, **headers)

        # Create a cache key for the client
        client_key = ClientCacheKey(
            url=url,
            headers=headers,
            timeout=self.timeout,
        )

        # Get or create the GraphQL client
        http_client = self._gql_client_cache.get(client_key)
        if http_client is None:
            transport = AIOHTTPTransport(
                url=url,
                headers=headers,
                timeout=self.timeout,
            )
            http_client = Client(transport=transport)
            self._gql_client_cache[client_key] = http_client
            self._transport_cache[client_key] = transport
            logger.debug(f"Created new GraphQL client for {url}")
        else:
            logger.debug(f"Reusing cached GraphQL client for {url}")

        query_gql = gql(query)
        query_gql.variable_values = variables or {}
        query_gql.operation_name = operation_name

        try:
            result: ExecutionResult = http_client.execute(
                query_gql,
                get_execution_result=True,
                upload_files=has_files,
            )
        except (TransportQueryError, TransportServerError) as e:
            logger.debug(f"GraphQL request failed: {e}", exc_info=True)
            raise

        return GraphQLResponseLike(result=result)

    def start_subscription(
        self, url: str, query: str, variables: dict, operation_name: str
    ) -> None:
        """Start a GraphQL subscription over WS using gql WebSockets transport.

        Args:
            url: The GraphQL WebSocket endpoint URL.
            query: The GraphQL subscription query string.
            variables: Variables for the subscription query.
            operation_name: Name of the subscription operation.

        Raises:
            ValueError: If operation_name is not provided.
            Exception: If the subscription fails to start.
        """
        if operation_name is None:
            raise ValueError("operation_name required for subscriptions")

        if operation_name in self._subscriptions:
            raise ValueError(
                f"Subscription with name '{operation_name}' already exists"
            )

        # Prepare headers
        headers = dict(self.default_headers)

        # Create WebSocket transport
        ws_url = url.replace("http://", "ws://").replace("https://", "wss://")
        ws_transport = WebsocketsTransport(
            url=ws_url,
            headers=headers,
            connect_timeout=self.timeout,
        )

        # Create client with WebSocket transport
        ws_client = Client(transport=ws_transport)

        # Parse the GraphQL query
        query_gql = gql(query)
        query_gql.variable_values = variables or {}
        query_gql.operation_name = operation_name

        # Execute the subscription - this returns a generator
        try:

            async def subscribe_async_wrapper() -> AsyncGenerator:
                """Wrapper for subscribe_async because that method does not close subscriptions properly."""
                async with ws_client as session:
                    generator = session.subscribe(query_gql)

                    self._to_close.append(generator)

                    async for result in generator:
                        yield result

            # Using the subscription as a generator that yields results
            # Run the async subscription in the event loop
            async def _create_subscription():
                self._subscriptions[operation_name] = subscribe_async_wrapper()

            self._threaded_async_loop.run_coroutine(
                _create_subscription(),
                timeout=5.0,
            )

            logger.debug(f"Started subscription {operation_name}")
        except Exception as e:
            logger.error(f"Failed to start subscription: {e}")
            raise

    async def get_next_message(
        self, op_name: str, timeout: float = 5.0
    ) -> Optional[_SubResponse]:
        """Get next message from subscription generator.

        Args:
            op_name: The name of the subscription operation.
            timeout: Timeout in seconds to wait for the next message.

        Returns:
            The next message from the subscription, or None if the subscription
            ended.

        Raises:
            ValueError: If the subscription name is not found.
            TimeoutError: If the timeout is reached while waiting.
            Exception: If an error occurs while getting the next message.
        """
        if op_name not in self._subscriptions:
            raise ValueError(
                f"Subscription with name '{op_name}' not found (have: {self._subscriptions.keys()} )"
            )

        subscription_generator = self._subscriptions[op_name]

        # Get the next item from the async iterator
        try:
            return await asyncio.wait_for(
                anext(subscription_generator), timeout=timeout
            )
        except StopAsyncIteration:
            logger.error(
                f"got unexpected StopAsyncIteration from subscription {op_name}"
            )
            return None
        except TimeoutError as e:
            logger.error(f"Timeout getting next message from subscription: {e}")
            raise
        except Exception as e:
            logger.error(f"Error getting next message from subscription: {e}")
            raise



from unittest.mock import Mock, patch

import pytest
from gql import FileVar

from tavern._core import exceptions
from tavern._plugins.graphql.client import GraphQLClient, GraphQLResponseLike
from tavern._plugins.graphql.request import GraphQLRequest, get_file_arguments


class TestGraphQLRequest:
    def test_init_valid_request(self, graphql_test_block_config):
        session = GraphQLClient()
        rspec = {
            "url": "http://example.com/graphql",
            "query": "query { hello }",
            "variables": {"name": "world"},
        }

        request = GraphQLRequest(session, rspec, graphql_test_block_config)

        assert request.request_vars.url == "http://example.com/graphql"
        assert request.request_vars.query == "query { hello }"
        assert request.request_vars.variables == {"name": "world"}

    def test_init_missing_query(self, graphql_test_block_config):
        session = GraphQLClient()
        rspec = {"url": "http://example.com/graphql", "variables": {"name": "world"}}

        with pytest.raises(
            exceptions.MissingKeysError,
            match="GraphQL request must contain 'query' field",
        ):
            GraphQLRequest(session, rspec, graphql_test_block_config)

    def test_init_missing_url(self, graphql_test_block_config):
        session = GraphQLClient()
        rspec = {"query": "query { hello }", "variables": {"name": "world"}}

        with pytest.raises(
            exceptions.MissingKeysError,
            match="GraphQL request must contain 'url' field",
        ):
            GraphQLRequest(session, rspec, graphql_test_block_config)

    def test_request_vars_defaults(self, graphql_test_block_config):
        session = GraphQLClient()
        rspec = {"url": "http://example.com/graphql", "query": "query { hello }"}

        request = GraphQLRequest(session, rspec, graphql_test_block_config)

        assert request.request_vars.variables == {}
        assert request.request_vars.operation_name is None
        assert request.request_vars.headers == {}

    def test_run_success(self, graphql_test_block_config):
        with (
            patch.object(GraphQLClient, "make_request") as mock_make_request,
        ):
            mock_response = Mock(spec=GraphQLResponseLike)
            mock_response.text = '{"data": {"hello": "world"}}'
            mock_make_request.return_value = mock_response

            session = GraphQLClient()
            rspec = {
                "url": "http://example.com/graphql",
                "query": "query { hello }",
                "headers": {"Authorization": "Bearer token"},
            }

            request = GraphQLRequest(session, rspec, graphql_test_block_config)
            response = request.run()

            mock_make_request.assert_called_once_with(
                url="http://example.com/graphql",
                query="query { hello }",
                variables={},
                operation_name=None,
                headers={
                    "Content-Type": "application/json",
                    "Authorization": "Bearer token",
                },
                has_files=False,
            )
            assert response.text == '{"data": {"hello": "world"}}'

    def test_run_failure(self, graphql_test_block_config):
        with patch.object(GraphQLClient, "make_request") as mock_make_request:
            mock_make_request.side_effect = Exception("Connection error")

            session = GraphQLClient()
            rspec = {"url": "http://example.com/graphql", "query": "query { hello }"}

            request = GraphQLRequest(session, rspec, graphql_test_block_config)

            with pytest.raises(
                exceptions.TavernException,
                match="GraphQL request failed: Connection error",
            ):
                request.run()


class TestGraphQLFileUploads:
    """Tests for GraphQL file upload functionality"""

    def test_get_file_arguments_single_file(self, graphql_test_block_config, tmp_path):
        """Test get_file_arguments with a single file"""
        # Create a test file
        test_file = tmp_path / "test.txt"
        test_file.write_text("test content")

        file_args = {"files": str(test_file)}

        result = get_file_arguments(file_args, graphql_test_block_config)

        # Should return a dict mapping 'files' to a FileVar
        assert "files" in result
        assert isinstance(result["files"], FileVar)

    def test_get_file_arguments_multiple_files(
        self, graphql_test_block_config, tmp_path
    ):
        """Test get_file_arguments with multiple files"""
        # Create test files
        test_file1 = tmp_path / "test1.txt"
        test_file1.write_text("test content 1")
        test_file2 = tmp_path / "test2.txt"
        test_file2.write_text("test content 2")

        file_args = {
            "file1": str(test_file1),
            "file2": str(test_file2),
        }

        result = get_file_arguments(file_args, graphql_test_block_config)

        # Should return a dict mapping variable names to FileVar objects
        assert "file1" in result
        assert "file2" in result
        assert isinstance(result["file1"], FileVar)
        assert isinstance(result["file2"], FileVar)

    def test_get_file_arguments_with_content_type(
        self, graphql_test_block_config, tmp_path
    ):
        """Test get_file_arguments with content type specified"""
        # Create a test file
        test_file = tmp_path / "test.json"
        test_file.write_text('{"key": "value"}')

        file_args = {
            "files": {
                "file_path": str(test_file),
                "content_type": "application/json",
            }
        }

        result = get_file_arguments(file_args, graphql_test_block_config)

        # Should return a dict mapping 'files' to a FileVar
        assert "files" in result
        assert isinstance(result["files"], FileVar)

    def test_get_file_arguments_empty_dict(self, graphql_test_block_config):
        """Test get_file_arguments with empty dict"""
        file_args = {}

        result = get_file_arguments(file_args, graphql_test_block_config)

        # Should return empty dict when no files
        assert result == {}

    def test_run_with_files_parameter(self, graphql_test_block_config, tmp_path):
        """Test running a GraphQL request with files parameter"""
        # Create a test file
        test_file = tmp_path / "test.txt"
        test_file.write_text("test content")

        with (
            patch.object(GraphQLClient, "make_request") as mock_make_request,
        ):
            mock_response = Mock(spec=GraphQLResponseLike)
            mock_response.text = '{"data": {"upload": "success"}}'
            mock_make_request.return_value = mock_response

            session = GraphQLClient()
            rspec = {
                "url": "http://example.com/graphql",
                "query": "mutation($files: Upload!) { singleUpload(file: $files) { id } }",
                "files": {"files": str(test_file)},
            }

            request = GraphQLRequest(session, rspec, graphql_test_block_config)
            response = request.run()

            # Verify that make_request was called
            mock_make_request.assert_called_once()

            # Get the call arguments
            call_args = mock_make_request.call_args

            # Check that variables include the files
            assert "variables" in call_args.kwargs
            variables = call_args.kwargs["variables"]
            assert "files" in variables
            assert isinstance(variables["files"], FileVar)

            assert response.text == '{"data": {"upload": "success"}}'

    def test_run_with_files_and_variables(self, graphql_test_block_config, tmp_path):
        """Test running a GraphQL request with both files and variables"""
        # Create a test file
        test_file = tmp_path / "test.txt"
        test_file.write_text("test content")

        with (
            patch.object(GraphQLClient, "make_request") as mock_make_request,
        ):
            mock_response = Mock(spec=GraphQLResponseLike)
            mock_response.text = '{"data": {"upload": "success"}}'
            mock_make_request.return_value = mock_response

            session = GraphQLClient()
            rspec = {
                "url": "http://example.com/graphql",
                "query": "mutation($file: Upload!, $title: String!) { uploadFile(file: $file, title: $title) { id } }",
                "files": {"file": str(test_file)},
                "variables": {"title": "My Upload"},
            }

            request = GraphQLRequest(session, rspec, graphql_test_block_config)
            response = request.run()

            # Verify that make_request was called
            mock_make_request.assert_called_once()

            # Get the call arguments
            call_args = mock_make_request.call_args

            # Check that variables include both the files and the regular variables
            assert "variables" in call_args.kwargs
            variables = call_args.kwargs["variables"]
            assert "file" in variables
            assert "title" in variables
            assert variables["title"] == "My Upload"
            assert isinstance(variables["file"], FileVar)

            assert response.text == '{"data": {"upload": "success"}}'

    def test_run_with_multiple_files_and_metadata(
        self, graphql_test_block_config, tmp_path
    ):
        """Test running a GraphQL request with multiple files including metadata"""
        # Create test files
        test_file1 = tmp_path / "test1.txt"
        test_file1.write_text("test content 1")
        test_file2 = tmp_path / "test2.json"
        test_file2.write_text('{"key": "value"}')

        with (
            patch.object(GraphQLClient, "make_request") as mock_make_request,
        ):
            mock_response = Mock(spec=GraphQLResponseLike)
            mock_response.text = '{"data": {"upload": "success"}}'
            mock_make_request.return_value = mock_response

            session = GraphQLClient()
            rspec = {
                "url": "http://example.com/graphql",
                "query": "mutation($file1: Upload!, $file2: Upload!) { multipleUpload(file1: $file1, file2: $file2) { id } }",
                "files": {
                    "file1": str(test_file1),
                    "file2": {
                        "file_path": str(test_file2),
                        "content_type": "application/json",
                    },
                },
            }

            request = GraphQLRequest(session, rspec, graphql_test_block_config)
            response = request.run()

            # Verify that make_request was called
            mock_make_request.assert_called_once()

            # Get the call arguments
            call_args = mock_make_request.call_args

            # Check that variables include the files
            assert "variables" in call_args.kwargs
            variables = call_args.kwargs["variables"]
            assert "file1" in variables
            assert "file2" in variables
            assert isinstance(variables["file1"], FileVar)
            assert isinstance(variables["file2"], FileVar)

            assert response.text == '{"data": {"upload": "success"}}'



import pytest


@pytest.fixture(scope="function", autouse=True)
def reset_db():
    """Reset the database between tests"""
    yield
    import requests

    requests.post("http://localhost:5010/reset")



#!/usr/bin/env python3

import re
import subprocess


def get_tags():
    """Get git tags sorted by creator date, filtered to exclude certain patterns."""
    result = subprocess.run(
        ["git", "tag", "--sort=creatordate"], capture_output=True, text=True, check=True
    )
    tags = result.stdout.strip().split("\n")
    # Filter out prerelease tags according to PEP 440
    # Matches versions with a/alpha, b/beta, or rc suffixes
    return [
        tag
        for tag in tags
        if not re.search(r"(a|alpha|b|beta|rc)\d*$", tag, re.IGNORECASE)
    ]


def get_tag_date(tag):
    """Get the date of a tag."""
    result = subprocess.run(
        ["git", "log", "-1", "--pretty=format:%ad", "--date=short", tag],
        capture_output=True,
        text=True,
        check=True,
    )
    return result.stdout.strip()


def get_tag_name(tag):
    """Get tag name with message."""
    result = subprocess.run(
        ["git", "tag", "-n1", tag], capture_output=True, text=True, check=True
    )
    return result.stdout.strip()


def get_tag_content(tag):
    """Get tag content (excluding first line)."""
    result = subprocess.run(
        ["git", "tag", "-n9", tag], capture_output=True, text=True, check=True
    )
    lines = result.stdout.strip().split("\n")
    if len(lines) > 1:
        content = "\n".join(lines[1:])
        # Remove leading whitespace from each line
        return "\n".join(line.lstrip() for line in content.split("\n"))
    return ""


def print_changelog():
    """Generate changelog content."""
    tags = get_tags()

    # Reverse to put oldest at bottom
    tags.reverse()

    output = ["# Changelog", ""]

    for current_tag in tags:
        tag_date = get_tag_date(current_tag)

        # Determine header level - minor releases (patch == 0) get '#', patch releases get '##'
        hashes = "##"
        parts = current_tag.split(".")
        if len(parts) >= 3 and parts[2] == "0":
            hashes = "#"

        tag_name = get_tag_name(current_tag)
        output.append(f"{hashes} {tag_name} ({tag_date})")

        tag_content = get_tag_content(current_tag)
        if tag_content:
            output.append("")
            output.append(tag_content)

        output.append("")

    return "\n".join(output)


if __name__ == "__main__":
    changelog = print_changelog()
    with open("CHANGELOG.md", "w") as f:
        f.write(changelog)



from unittest.mock import Mock

import requests

from tavern._plugins.graphql.client import GraphQLClient
from tavern._plugins.graphql.response import GraphQLResponse


class TestGraphQLResponse:
    def test_init_defaults(self, graphql_test_block_config):
        session = Mock(spec=GraphQLClient)
        expected = {}

        response = GraphQLResponse(session, "test", expected, graphql_test_block_config)

        assert response.expected["status_code"] == 200

    def test_init_custom_status_code(self, graphql_test_block_config):
        session = Mock(spec=GraphQLClient)
        expected = {"status_code": 201}

        response = GraphQLResponse(session, "test", expected, graphql_test_block_config)

        assert response.expected["status_code"] == 201

    def test_init_list_status_code(self, graphql_test_block_config):
        session = Mock(spec=GraphQLClient)
        expected = {"status_code": [200, 201]}

        response = GraphQLResponse(session, "test", expected, graphql_test_block_config)

        assert response.expected["status_code"] == [200, 201]

    def test_str_with_response(self, graphql_test_block_config):
        session = Mock(spec=GraphQLClient)
        expected = {}

        mock_response = Mock(spec=requests.Response)
        mock_response.text = '{"data": {"hello": "world"}}'

        response = GraphQLResponse(session, "test", expected, graphql_test_block_config)
        response.response = mock_response

        assert str(response) == '{"data": {"hello": "world"}}'

    def test_str_without_response(self, graphql_test_block_config):
        session = Mock(spec=GraphQLClient)
        expected = {}

        response = GraphQLResponse(session, "test", expected, graphql_test_block_config)

        assert str(response) == "<Not run yet>"

    def test_validate_response_format_valid_data(self, graphql_test_block_config):
        session = Mock(spec=GraphQLClient)
        expected = {}

        response = GraphQLResponse(session, "test", expected, graphql_test_block_config)

        # Should not raise any exception
        response._validate_graphql_response_structure({"data": {"hello": "world"}})
        assert len(response.errors) == 0

    def test_validate_response_format_valid_errors(self, graphql_test_block_config):
        session = Mock(spec=GraphQLClient)
        expected = {}

        response = GraphQLResponse(session, "test", expected, graphql_test_block_config)

        # Should not raise any exception, just log warning
        response._validate_graphql_response_structure(
            {"errors": [{"message": "Something went wrong"}]}
        )
        assert len(response.errors) == 0

    def test_validate_response_format_invalid_no_data_or_errors(
        self, graphql_test_block_config
    ):
        session = Mock(spec=GraphQLClient)
        expected = {}

        response = GraphQLResponse(session, "test", expected, graphql_test_block_config)

        response._validate_graphql_response_structure({"other": "field"})
        assert len(response.errors) == 2
        assert (
            "Response must contain either 'data' or 'errors' at the top level"
            in response.errors[0]
        )
        assert (
            "Invalid GraphQL top-level keys: {'other'}. Only 'data' and 'errors' are allowed"
            in response.errors[1]
        )



"""VERY simple skeleton for plugin stuff

This is here mainly to make MQTT easier, this will almost defintiely change
significantly if/when a proper plugin system is implemented!
"""

import dataclasses
import logging
from collections.abc import Callable, Mapping
from functools import partial
from typing import Any, Optional, Protocol

import stevedore
import stevedore.extension

from tavern._core import exceptions
from tavern._core.dict_util import format_keys
from tavern._core.pytest.config import TestConfig
from tavern.request import BaseRequest
from tavern.response import BaseResponse

logger: logging.Logger = logging.getLogger(__name__)


class PluginHelperBase:
    """Base for plugins"""


def plugin_load_error(mgr, entry_point, err):
    """Handle import errors"""
    msg = f"Error loading plugin {entry_point} - {err}"
    raise exceptions.PluginLoadError(msg) from err


class _TavernPlugin(Protocol):
    """A tavern plugin"""

    session_type: type[Any]
    request_type: type[BaseRequest]
    verifier_type: type[BaseResponse]
    response_block_name: str
    request_block_name: str
    schema: Mapping
    has_multiple_responses: bool

    def get_expected_from_request(
        self, response_block: BaseResponse, test_block_config: TestConfig, session: Any
    ) -> Any: ...


def is_valid_reqresp_plugin(ext: stevedore.extension.Extension) -> bool:
    """Whether this is a valid 'reqresp' plugin

    Requires certain functions/variables to be present

    Todo:
        Not all of these are required for all request/response types probably

    Args:
        ext: class or module plugin object

    Returns:
        Whether the plugin has everything we need to use it
    """
    required = [
        # MQTTClient, requests.Session
        "session_type",
        # RestRequest, MQTTRequest
        "request_type",
        # request, mqtt_publish, grpc_request
        "request_block_name",
        # Some function that returns a dict
        "get_expected_from_request",
        # MQTTResponse, RestResponse
        "verifier_type",
        # response, mqtt_response, grpc_request
        "response_block_name",
        # dictionary with pykwalify schema
        "schema",
        # whether plugin supports multiple responses (e.g., graphql_responses, mqtt_responses)
        "has_multiple_responses",
    ]

    plugin: _TavernPlugin = ext.plugin

    return all(hasattr(plugin, i) for i in required)


class _Plugin:
    """Wrapped tavern plugin for convenience"""

    name: str
    plugin: _TavernPlugin


@dataclasses.dataclass
class _PluginCache:
    plugins: list[_Plugin] = dataclasses.field(default_factory=list)

    def __call__(self, config: Optional[TestConfig] = None) -> list[_Plugin]:
        if self.plugins:
            return self.plugins

        if config:
            # NOTE: This is reloaded every time
            self.plugins = self._load_plugins(config)
            return self.plugins

        raise exceptions.PluginLoadError("No config to load plugins from")

    def _load_plugins(self, test_block_config: TestConfig) -> list[_Plugin]:
        """Load plugins from the 'tavern' entrypoint namespace

        This can be a module or a class as long as it defines the right things

        Todo:
            - Limit which plugins are loaded based on some config/command line
              option
            - Different plugin names

        Args:
            test_block_config: available config for test

        Raises:
            exceptions.MissingSettingsError: invalid entry points set

        Returns:
            Loaded plugins, can be a class or a module
        """

        plugins = []
        discovered_plugins: dict[str, list[str]] = {}

        def is_plugin_backend_enabled(
            current_backend: str, ext: stevedore.extension.Extension
        ) -> bool:
            """Checks if a plugin backend is enabled based on configuration.

            If no specific backend is configured, defaults to enabled.
            Adds enabled plugins to discovered_plugins tracking dictionary.

            Args:
                current_backend: The backend being checked (e.g. 'http', 'mqtt')
                ext: The stevedore extension object representing the plugin

            Returns:
                Whether the plugin backend is enabled
            """
            if test_block_config.tavern_internal.backends[current_backend] is None:
                # Use whatever default - will raise an error if >1 is discovered
                is_enabled = True
                logger.debug(f"Using default backend for {ext.name}")
            else:
                is_enabled = (
                    ext.name
                    == test_block_config.tavern_internal.backends[current_backend]
                )
                logger.debug(
                    f"Is {current_backend} for {ext.name} enabled? {is_enabled}"
                )

            if is_enabled:
                if current_backend not in discovered_plugins:
                    discovered_plugins[current_backend] = []
                discovered_plugins[current_backend].append(ext.name)

            return is_enabled

        for backend in test_block_config.tavern_internal.backends.keys():
            logger.debug("loading backend for %s", backend)

            namespace = f"tavern_{backend}"

            manager = stevedore.EnabledExtensionManager(
                namespace=namespace,
                check_func=partial(is_plugin_backend_enabled, backend),
                verify_requirements=True,
                on_load_failure_callback=plugin_load_error,
            )
            manager.propagate_map_exceptions = True

            validation_results = manager.map(is_valid_reqresp_plugin)
            invalid_plugins = [
                ext.name
                for ext, is_valid in zip(manager.extensions, validation_results)
                if not is_valid
            ]
            if invalid_plugins:
                raise exceptions.PluginLoadError(
                    f"Plugin(s) {invalid_plugins} failed validation: "
                    f"missing required 'has_multiple_responses' field or other required attributes. "
                    f"Required fields: session_type, request_type, request_block_name, "
                    f"get_expected_from_request, verifier_type, response_block_name, schema, "
                    f"has_multiple_responses"
                )

            if len(manager.extensions) != 1:
                raise exceptions.MissingSettingsError(
                    f"Expected exactly one entrypoint in 'tavern-{backend}' namespace but got {len(manager.extensions)}"
                )

            plugins.extend(manager.extensions)

        for plugin, enabled in discovered_plugins.items():
            if len(enabled) > 1:
                raise exceptions.PluginLoadError(
                    f"Multiple plugins enabled for '{plugin}' backend: {enabled}"
                )

        return plugins


load_plugins = _PluginCache()


def get_extra_sessions(test_spec: Mapping, test_block_config: TestConfig) -> dict:
    """Get extra 'sessions' for any extra test types

    Args:
        test_spec: Spec for the test block
        test_block_config: available config for test

    Returns:
        mapping of name to session. Session should be a context manager.
    """

    sessions = {}

    plugins: list[_Plugin] = load_plugins(test_block_config)

    for p in plugins:
        if any(
            (p.plugin.request_block_name in i or p.plugin.response_block_name in i)
            for i in test_spec["stages"]
        ):
            logger.debug(
                "Initialising session for %s (%s)", p.name, p.plugin.session_type
            )
            session_spec: dict = test_spec.get(p.name, {})
            formatted: dict = format_keys(session_spec, test_block_config.variables)
            sessions[p.name] = p.plugin.session_type(**formatted)

    return sessions


def get_request_type(
    stage: Mapping,
    test_block_config: TestConfig,
    sessions: Mapping,
) -> BaseRequest:
    """Get the request object for this stage

    there can only be one

    Args:
        stage: spec for this stage
        test_block_config: variables for this test run
        sessions: all available sessions

    Returns:
        request object with a run() method

    Raises:
        exceptions.DuplicateKeysError: More than one kind of request specified
        exceptions.MissingKeysError: No request type specified
    """

    plugins = load_plugins(test_block_config)

    keys = {}

    for p in plugins:
        keys[p.plugin.request_block_name] = p.plugin.request_type

    if len(set(keys) & set(stage)) > 1:
        raise exceptions.DuplicateKeysError(
            f"Can only specify 1 request type but got {set(keys)}"
        )
    elif not list(set(keys) & set(stage)):
        raise exceptions.MissingKeysError(
            f"Need to specify one of valid request types: '{set(keys.keys())}'"
        )

    # We've validated that 1 and only 1 is there, so just loop until the first
    # one is found
    request_class: type[BaseRequest] | None = None
    for p in plugins:
        try:
            request_args = stage[p.plugin.request_block_name]
        except KeyError:
            pass
        else:
            session = sessions[p.name]
            request_class = p.plugin.request_type
            logger.debug(
                "Initialising request class for %s (%s)", p.name, request_class
            )
            break

    if not request_class:
        raise exceptions.MissingSettingsError("No request type found")

    request_maker = request_class(session, request_args, test_block_config)

    return request_maker


class ResponseVerifier(dict):
    plugin_name: str


def _foreach_response(
    stage: Mapping,
    test_block_config: TestConfig,
    action: Callable[[_Plugin, str], dict],
) -> dict[str, dict]:
    """Do something for each response

    Args:
        stage: Stage of test
        test_block_config: Config for test
        action: function that takes (plugin, response block)

    Returns:
        mapping of plugin name to list of expected (normally length 1)
    """

    plugins = load_plugins(test_block_config)

    retvals = {}

    for p in plugins:
        response_block = stage.get(p.plugin.response_block_name)
        if response_block is not None:
            retvals[p.name] = action(p, response_block)

    return retvals


def get_expected(
    stage: Mapping,
    test_block_config: TestConfig,
    sessions: Mapping,
):
    """Get expected responses for each type of request

    Though only 1 request can be made, it can cause multiple responses.

    Because we need to subcribe to MQTT topics, which might be formatted from
    keys from included files, the 'expected'/'response' needs to be formatted
    BEFORE running the request.

    Args:
        stage: test stage
        test_block_config: available configuration for this test
        sessions: all available sessions

    Returns:
        mapping of request type to expected response dict
    """

    def action(p, response_block):
        plugin_expected = p.plugin.get_expected_from_request(
            response_block, test_block_config, sessions[p.name]
        )
        if plugin_expected:
            plugin_expected = ResponseVerifier(**plugin_expected)
            plugin_expected.plugin_name = p.name
            return plugin_expected
        else:
            return None

    return _foreach_response(stage, test_block_config, action)


def get_verifiers(
    stage: Mapping,
    test_block_config: TestConfig,
    sessions: Mapping,
    expected: Mapping,
):
    """Get one or more response validators for this stage

    Args:
        stage: spec for this stage
        test_block_config: variables for this test run
        sessions: all available sessions
        expected: expected responses for this stage

    Returns:
        response validator object with a verify(response) method
    """

    def action(p, _):
        session = sessions[p.name]
        logger.debug(
            "Initialising verifier for %s (%s)", p.name, p.plugin.verifier_type
        )
        verifiers = []

        plugin_expected = expected[p.name]

        verifier = p.plugin.verifier_type(
            session,
            stage["name"],
            plugin_expected,
            test_block_config,
        )
        verifiers.append(verifier)

        return verifiers

    return _foreach_response(stage, test_block_config, action)



import logging
from typing import Any

from box.box import Box

from tavern._core import exceptions
from tavern._core.pytest import call_hook
from tavern._core.report import attach_yaml
from tavern._plugins.common.response import CommonResponse
from tavern._plugins.graphql.client import (
    GraphQLClient,
    GraphQLResponseLike,
    ResultOrErr,
)

logger: logging.Logger = logging.getLogger(__name__)


class GraphQLResponse(CommonResponse):
    """GraphQL response verification for HTTP and WS"""

    session: GraphQLClient

    def __init__(
        self,
        session,
        name: str,
        expected: dict,
        test_block_config,
    ):
        """Initialize GraphQL response validator.

        Args:
            session: GraphQL client instance
            name: Name of the test block
            expected: Expected response configuration
            test_block_config: Test block configuration
        """
        self.session = session

        sync_responses = 0
        for resp in expected.get("graphql_responses", []):
            if resp.get("subscription") is None:
                sync_responses += 1

        if sync_responses > 1:
            raise exceptions.BadSchemaError(
                "Only one graphql_response can be synchronous"
            )

        expected["save"] = expected.get("save", {})
        for e in expected.get("graphql_responses", []):
            save_block: dict
            if save_block := e.get("save", {}):
                if not isinstance(save_block, dict):
                    raise exceptions.BadSchemaError(
                        "save block for graphql_response must be a dict"
                    )

                for to_save in save_block:
                    if to_save in expected["save"]:
                        raise exceptions.BadSchemaError(
                            f"save block for graphql_response cannot contain duplicate keys: {to_save}"
                        )

                expected["save"].update(save_block)

        super().__init__(
            session,
            name,
            expected,
            test_block_config,
            multiple_responses_block="graphql_responses",
        )

    def _validate_graphql_response_structure(self, body: Any) -> None:
        """Validate GraphQL response structure: data or errors top-level

        Checks if the response body contains valid GraphQL response structure
        with either 'data' or 'errors' keys at the top level.

        Args:
            body: Response body to validate
        """
        if body is None:
            self._adderr("GraphQL response body missing")
            return
        if not isinstance(body, dict):
            self._adderr("GraphQL body not dict: %r", body)
            return

        allowed = {"data", "errors"}
        if not allowed & set(body):
            self._adderr(
                "Response must contain either 'data' or 'errors' at the top level"
            )
        if set(body) - allowed:
            self._adderr(
                "Invalid GraphQL top-level keys: %s. Only 'data' and 'errors' are allowed",
                set(body) - allowed,
            )

    def verify(self, response: GraphQLResponseLike) -> dict:
        """Verify response against expected values and returns any values that
        we wanted to save for use in future requests.

        Args:
            response: The GraphQL response to verify

        Returns:
            Dictionary of saved values from the response

        Raises:
            TestFailError: If verification fails with collected errors
        """

        graphql_responses = self.expected.get("graphql_responses", [])

        sync_responses_list: list[dict] = [
            resp for resp in graphql_responses if "subscription" not in resp
        ]

        # Process synchronous responses (will actually be length 1)
        saved = {}
        for expected_resp in sync_responses_list:
            saved.update(self._check_sync_response(expected_resp, response))

        sub_responses_list = [
            resp for resp in graphql_responses if "subscription" in resp
        ]
        # Process subscription responses
        if sub_responses_list:
            try:
                to_save: dict = self.session._threaded_async_loop.run_coroutine(
                    self._handle_subscription_responses(sub_responses_list),
                    timeout=30.0,
                )
            except TimeoutError as e:
                self._adderr(f"Timed out waiting for subscription responses: {e}")
            else:
                saved.update(to_save)

        if self.errors:
            raise exceptions.TestFailError(
                f"Test '{self.name:s}' failed:\n{self._str_errors():s}",
                failures=self.errors,
            )

        return saved

    async def _handle_subscription_responses(
        self, sub_responses_list: list[dict]
    ) -> dict:
        """Handle subscription responses concurrently.

        Processes subscription-based GraphQL responses by waiting for messages
        on each subscription and validating them against expected responses.

        Args:
            sub_responses_list: List of subscription response configurations

        Returns:
            Dictionary of saved values from subscription responses
        """

        async def get_subscription_results(
            expected_resp,
        ) -> ResultOrErr | None:
            """Get subscription message result for an expected response.

            Waits for the next message on a subscription operation and returns
            the result.

            Args:
                expected_resp: Expected response configuration for subscription

            Returns:
                response or None if error occurred
            """
            op_name = expected_resp["subscription"]
            timeout: int | float = expected_resp.get("timeout", 3.0)
            try:
                response = await self.session.get_next_message(op_name, timeout)
            except TimeoutError:
                self._adderr(
                    f"Timed out waiting for subscription message on '{op_name}'"
                )
                return None
            except Exception as e:
                logger.exception(
                    f"Error getting subscription message on '{op_name}': {e}"
                )
                self._adderr(f"Error getting subscription message on '{op_name}': {e}")
                return None

            if response is None:
                self._adderr(f"Subscription message on '{op_name}' was None")
                return None

            return response

        saved = {}

        async def async_generator_wrapper():
            """Asynchronous generator that yields expected responses with results.

            For each expected subscription response, gets the subscription result
            and yields the pair for processing.

            Yields:
                Tuple of (expected_resp, response) for each subscription
            """
            for resp in sub_responses_list:
                yield resp, await get_subscription_results(resp)

        # Process the results using async for
        async for expected_resp, response in async_generator_wrapper():
            if response is None:
                continue
            op_name = expected_resp["subscription"]
            # Process successful subscription response
            body = response
            if "json" in expected_resp:
                self._validate_block("json", body)

            attach_yaml(
                {"ws_op": op_name, "body": body},
                name="graphql_ws_response",
            )

            call_hook(
                self.test_block_config,
                "pytest_tavern_beta_after_every_response",
                expected=expected_resp,
                response=response,
            )
            self._maybe_run_validate_functions(response)

            saved.update(self._common_verify_save(body, response))
            saved.update(
                self.maybe_get_save_values_from_save_block("data", {"data": body})
            )

        return saved

    def _check_sync_response(
        self,
        expected_resp: dict,
        response: GraphQLResponseLike,
    ) -> dict[Any, Any]:
        """Check a synchronous GraphQL response against expected values.

        Validates a synchronous (non-subscription) GraphQL response by checking
        for expected errors and validating the response data.

        Args:
            expected_resp: Expected response configuration
            response: Actual GraphQL response to validate

        Returns:
            Dictionary of saved values from the response
        """
        call_hook(
            self.test_block_config,
            "pytest_tavern_beta_after_every_response",
            expected=expected_resp,
            response=response,
        )

        # Regular HTTP GraphQL response
        logger.info(f"response: {response}")

        expected_errors: list[str]
        if expected_errors := expected_resp.get("errors", []):
            if not response.result.errors:
                self._adderr("Expected errors but got none")
                return {}

            if len(expected_errors) != len(response.result.errors):
                self._adderr(
                    f"Expected {len(expected_errors)} errors but got {len(response.result.errors)}"
                )
                # Continue and do a best effort check

            got_error_messages = [
                Box(error).message for error in response.result.errors
            ]
            expected_errors = [Box(error).message for error in expected_errors]
            for expected_error in expected_errors:
                found = False
                for got_error in got_error_messages:
                    if expected_error in got_error:
                        found = True

                if not found:
                    self._adderr(
                        f"error message '{expected_error}' not found in returned error messages (had {got_error_messages})"
                    )
        elif response.result.errors:
            self._adderr(
                f"got errors when none were expected: {response.result.errors}"
            )
            return {}

        body = response.result.data
        self._validate_block("data", body, read_from=expected_resp)  # type:ignore

        attach_yaml(
            {
                "body": body,
            },
            name="graphql_response",
        )

        self._maybe_run_validate_functions(response)

        saved = {}
        saved.update(self._common_verify_save(body, response))
        saved.update(self.maybe_get_save_values_from_save_block("data", {"data": body}))

        return saved



import logging
from collections.abc import Iterable
from os.path import abspath, dirname, join
from typing import Optional, Union

import yaml

from tavern._core.dict_util import format_keys
from tavern._core.pytest.config import TestConfig

from .client import GraphQLClient
from .request import GraphQLRequest
from .response import GraphQLResponse

logger: logging.Logger = logging.getLogger(__name__)


session_type = GraphQLClient
request_type = GraphQLRequest
request_block_name = "graphql_request"
verifier_type = GraphQLResponse
response_block_name = "graphql_response"
has_multiple_responses = True


def get_expected_from_request(
    response_block: Union[dict, Iterable[dict]],
    test_block_config: TestConfig,
    session: GraphQLClient,
) -> Optional[dict]:
    if response_block is None:
        return None

    expected: dict[str, list] = {"graphql_responses": []}
    if isinstance(response_block, dict):
        response_block = [response_block]

    for resp_block in response_block:
        f_expected = format_keys(resp_block, test_block_config.variables)
        expected["graphql_responses"].append(f_expected)

    return expected


# Schema validation
schema_path: str = join(abspath(dirname(__file__)), "jsonschema.yaml")
with open(schema_path, encoding="utf-8") as schema_file:
    schema = yaml.load(schema_file, Loader=yaml.SafeLoader)



import logging
import pathlib
from collections.abc import Iterable
from os.path import abspath, dirname, join
from typing import Any, Optional, Union

import box
import yaml

from tavern._core import exceptions
from tavern._core.pytest.config import TestConfig
from tavern.request import BaseRequest
from tavern.response import BaseResponse


class Session:
    """No-op session, but must implement the context manager protocol"""
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class Request(BaseRequest):
    """Touches a file when the 'request' is made"""
    def __init__(
        self, session: Any, rspec: dict, test_block_config: TestConfig
    ) -> None:
        self.session = session

        self._request_vars = rspec

    @property
    def request_vars(self) -> box.Box:
        return self._request_vars

    def run(self):
        pathlib.Path(self._request_vars["filename"]).touch()


class Response(BaseResponse):
    def verify(self, response):
        if not pathlib.Path(self.expected["filename"]).exists():
            raise exceptions.BadSchemaError(
                f"Expected file '{self.expected['filename']}' does not exist"
            )

        return {}

    def __init__(
        self,
        client,
        name: str,
        expected: TestConfig,
        test_block_config: TestConfig,
    ) -> None:
        super().__init__(name, expected, test_block_config)


logger: logging.Logger = logging.getLogger(__name__)

session_type = Session

request_type = Request
request_block_name = "touch_file"


verifier_type = Response
response_block_name = "file_exists"

has_multiple_responses = False


def get_expected_from_request(
    response_block: Union[dict, Iterable[dict]],
    test_block_config: TestConfig,
    session: Session,
) -> Optional[dict]:
    return response_block


schema_path: str = join(abspath(dirname(__file__)), "jsonschema.yaml")
with open(schema_path, encoding="utf-8") as schema_file:
    schema = yaml.load(schema_file, Loader=yaml.SafeLoader)



import logging
import logging.config

import pytest
import yaml

@pytest.fixture(autouse=True)
def setup_logging():
    log_cfg = """
---
version: 1
formatters:
    default:
        # colorlog is really useful
        (): colorlog.ColoredFormatter
        format: "%(asctime)s [%(bold)s%(log_color)s%(levelname)s%(reset)s]: (%(bold)s%(name)s:%(lineno)d%(reset)s) %(message)s"
        style: "%"
        datefmt: "%X"
        log_colors:
            DEBUG:    cyan
            INFO:     green
            WARNING:  yellow
            ERROR:    red
            CRITICAL: red,bg_white

handlers:
    stderr:
        class: colorlog.StreamHandler
        formatter: default

loggers:
    tavern:
        handlers:
            - stderr
        level: INFO
        propagate: false
"""

    as_dict = yaml.load(log_cfg, Loader=yaml.SafeLoader)
    logging.config.dictConfig(as_dict)

    logging.info("Logging set up")


"""Simple GraphQL test server for integration testing using SQLite and strawberry-graphql with subscriptions"""

import asyncio
import csv
import logging
from collections.abc import AsyncGenerator

from starlette.background import BackgroundTask
from starlette.responses import Response
from strawberry.file_uploads import Upload
import starlette
from starlette.datastructures import UploadFile
import starlette.authentication
import starlette.middleware.authentication
import starlette.websockets
import strawberry
import uvicorn
from fastapi import FastAPI
from sqlalchemy import Column, ForeignKey, Integer, String, create_engine, select, text
from sqlalchemy.orm import declarative_base, relationship, sessionmaker
from starlette.requests import HTTPConnection
from strawberry.exceptions import ConnectionRejectionError
from strawberry.fastapi import GraphQLRouter
from strawberry_sqlalchemy_mapper import StrawberrySQLAlchemyMapper

strawberry_sqlalchemy_mapper = StrawberrySQLAlchemyMapper()

Base = declarative_base()


class models:
    """dummy module to contains models so they dont have to be a in a separate file

    It seems like there is undocumented(?) feature of strawberry that the original
    model has to have the same name as the graphql type
    """

    class User(Base):
        __tablename__ = "users"

        id = Column(Integer, primary_key=True)
        name = Column(String(100), nullable=False)
        email = Column(String(100), nullable=False)

    class Post(Base):
        __tablename__ = "posts"

        id = Column(Integer, primary_key=True)
        title = Column(String(200), nullable=False)
        content = Column(String, nullable=False)
        author_id = Column(Integer, ForeignKey("users.id"), nullable=False)
        author = relationship("User", backref="posts")


@strawberry_sqlalchemy_mapper.type(models.User)
class User:
    pass


@strawberry_sqlalchemy_mapper.type(models.Post)
class Post:
    pass


@strawberry.type
class Query:
    @strawberry.field
    def user(self, id: strawberry.ID) -> User:
        user = global_db_session.get(models.User, int(id))
        if user is None:
            raise Exception("User not found")
        return user

    @strawberry.field
    def users(self) -> list[User]:
        return global_db_session.execute(select(models.User)).scalars().all()

    @strawberry.field
    def post(self, id: strawberry.ID) -> Post:
        post = global_db_session.get(models.Post, int(id))
        if post is None:
            raise Exception("Post not found")
        return post

    @strawberry.field
    def posts(self) -> list[Post]:
        return global_db_session.execute(select(models.Post)).scalars().all()

    @strawberry.field
    def user_posts(self, author_id: strawberry.ID) -> list[Post]:
        posts_by_author = list(
            global_db_session.execute(
                select(models.Post).filter_by(author_id=int(author_id))
            )
            .scalars()
            .all()
        )
        if not posts_by_author:
            raise Exception("No posts found for that author")
        return posts_by_author

    @strawberry.field
    async def users_from_csv(
        self, csv_file: Upload, info: strawberry.Info
    ) -> list[User]:
        content = (await csv_file.read()).decode("utf-8")
        csv_reader = csv.reader(content.splitlines())
        first_column_values = [row[0] for row in csv_reader if row]

        matching_users = []
        for name in first_column_values:
            user = global_db_session.execute(
                select(models.User).where(models.User.name == name)
            ).scalars().first()
            if user:
                matching_users.append(user)

        return matching_users


@strawberry.type
class Mutation:
    @strawberry.mutation
    def create_user(self, name: str, email: str, info: strawberry.Info) -> User:
        user = models.User(name=name, email=email)
        global_db_session.add(user)
        global_db_session.commit()
        q.put_nowait(user)
        return user

    @strawberry.mutation
    def create_post(
        self, title: str, content: str, author_id: str, info: strawberry.Info
    ) -> Post:
        post = models.Post(title=title, content=content, author_id=int(author_id))
        global_db_session.add(post)
        global_db_session.commit()
        return post

    @strawberry.mutation
    async def create_post_from_file(
        self, title: str, my_file_name: Upload, author_id: str, info: strawberry.Info
    ) -> Post:
        post = models.Post(
            title=title,
            content=(await my_file_name.read()).decode("utf-8"),
            author_id=int(author_id),
        )
        logging.info(f"Creating post {post.title}")
        global_db_session.add(post)
        global_db_session.commit()
        return post

    @strawberry.mutation
    def update_user(
        self, id: strawberry.ID, name: str, email: str, info: strawberry.Info
    ) -> User:
        user = global_db_session.get(models.User, int(id))
        if user is None:
            raise Exception("User not found")
        user.name = name
        user.email = email
        global_db_session.commit()
        q.put_nowait(user)
        return user


q = asyncio.Queue()


@strawberry.type
class Subscription:
    @strawberry.subscription(graphql_type=User)
    async def user(self, id: strawberry.ID) -> AsyncGenerator[User, None]:
        logging.info(f"starting subscription for user {id}")
        while True:
            user = await q.get()
            logging.info(f"User {user.name} updated")
            if user.id != int(id):
                continue
            yield user


strawberry_sqlalchemy_mapper.finalize()
schema = strawberry.Schema(
    query=Query,
    mutation=Mutation,
    subscription=Subscription,
    scalar_overrides={UploadFile: Upload},
)


async def lifespan(_):
    logging.basicConfig(level=logging.DEBUG)
    yield


app = FastAPI(title="GraphQL Test Server", lifespan=lifespan)


@app.middleware("http")
async def some_middleware(request, call_next):
    req_body = await request.body()
    # await set_body(request, req_body)  # not needed when using FastAPI>=0.108.0.
    response = await call_next(request)

    chunks = []
    async for chunk in response.body_iterator:
        chunks.append(chunk)
    res_body = b"".join(chunks)

    def log_info(req_body, res_body):
        logging.info(req_body)
        logging.info(res_body)

    task = BackgroundTask(log_info, req_body, res_body)
    return Response(
        content=res_body,
        status_code=response.status_code,
        headers=dict(response.headers),
        media_type=response.media_type,
        background=task,
    )


class BasicAuthBackend(starlette.authentication.AuthenticationBackend):
    async def authenticate(
        self, conn: HTTPConnection
    ) -> (
        tuple[
            starlette.authentication.AuthCredentials,
            starlette.authentication.SimpleUser,
        ]
        | None
    ):
        # Check if Authorization header exists
        headers = conn.headers
        return check_has_right_header(headers)


def check_has_right_header(headers):
    """Checks if the authorization header is set to the correct thing"""
    if "authorization" not in headers:
        logging.info("No authorization header found")
        return None

    auth_header = headers["authorization"]
    logging.info(f"Got authorization header: {auth_header}")

    try:
        scheme, credentials = auth_header.split()
        if scheme.lower() != "bearer":
            return None

        if credentials == "test-token":
            # Return authenticated user with credentials
            return starlette.authentication.AuthCredentials(
                ["authenticated"]
            ), starlette.authentication.SimpleUser("tavern")
    except Exception as e:
        logging.error(f"Authentication failed: {e}")
        raise

    return None


app.add_middleware(
    starlette.middleware.authentication.AuthenticationMiddleware,
    backend=BasicAuthBackend(),
)


class AuthenticatedGraphqlRouter(GraphQLRouter):
    async def on_ws_connect(self, context: dict[str, object]):
        request: starlette.websockets.WebSocket = context["request"]
        if request.auth is None:
            logging.info("No authentication found")
            raise ConnectionRejectionError({"message": "Unauthorized"})

        logging.info(f"Got authenticated request from {request.client.host}")

        # Accept without a acknowledgment payload
        return await super().on_ws_connect(context)

    def is_request_allowed(self, request):
        if not super().is_request_allowed(request):
            return False

        # This feels like a terrible hack but I dont think strawberry provides another way to do this
        if check_has_right_header(request.headers) is None:
            from cross_web import HTTPException

            raise HTTPException(401, "Unauthorized")

        return True


app.include_router(
    GraphQLRouter(schema, multipart_uploads_enabled=True), prefix="/graphql"
)
app.include_router(
    AuthenticatedGraphqlRouter(schema),
    prefix="/graphql_authenticated",
)


@app.post("/reset")
async def reset_db():
    global_db_session.execute(text("delete from users"))
    global_db_session.execute(text("delete from posts"))
    global_db_session.commit()
    while not q.empty():
        q.get_nowait()
    logging.info("Database reset")


@app.get("/health")
async def health():
    return {"status": "healthy"}


# DB setup
engine = create_engine("sqlite:////tmp/test.db", echo=False)
Session = sessionmaker(bind=engine)
global_db_session = Session()
Base.metadata.create_all(bind=engine)


if __name__ == "__main__":
    uvicorn.run(
        app,
        host="0.0.0.0",
        port=5010,
        log_level="info",
    )



import contextlib
import dataclasses
import pathlib
import tempfile
from collections.abc import Callable, Generator
from typing import Any
from unittest.mock import Mock

import pytest
import yaml

from tavern._core import exceptions
from tavern._core.pytest.file import YamlFile
from tavern._core.pytest.item import YamlItem


@pytest.fixture(scope="function")
def tavern_test_content():
    """return some example tests"""

    test_docs = [
        {"test_name": "First test", "stages": [{"name": "stage 1"}]},
        {"test_name": "Second test", "stages": [{"name": "stage 2"}]},
        {"test_name": "Third test", "stages": [{"name": "stage 3"}]},
    ]

    return test_docs


@contextlib.contextmanager
def tavern_test_file(test_content: list[Any]) -> Generator[pathlib.Path, Any, None]:
    """Create a temporary YAML file with multiple documents"""

    with tempfile.TemporaryDirectory() as tmpdir:
        file_path = pathlib.Path(tmpdir) / "test.yaml"

        # Write the documents to the file
        with file_path.open("w", encoding="utf-8") as f:
            for doc in test_content:
                yaml.dump(doc, f)
                f.write("---\n")

        yield file_path


@dataclasses.dataclass
class Opener:
    """Simple mock for generating items because pytest makes it hard to wrap
    their internal functionality"""

    path: pathlib.Path
    _generate_items: Callable[[dict], Any]


class TestGenerateFiles:
    @pytest.mark.parametrize("with_merge_down_test", (True, False))
    def test_multiple_documents(self, tavern_test_content, with_merge_down_test):
        """Verify that multiple documents in a YAML file result in multiple tests"""

        # Collect all tests
        if with_merge_down_test:
            tavern_test_content.insert(0, {"includes": [], "is_defaults": True})

        def generate_yamlitem(test_spec):
            mock = Mock(spec=YamlItem)
            mock.name = test_spec["test_name"]
            yield mock

        with tavern_test_file(tavern_test_content) as filename:
            tests = list(
                YamlFile.collect(
                    Opener(
                        path=filename,
                        _generate_items=generate_yamlitem,
                    )
                )
            )

        assert len(tests) == 3

        # Verify each test has the correct name
        expected_names = ["First test", "Second test", "Third test"]
        for test, expected_name in zip(tests, expected_names):
            assert test.name == expected_name

    @pytest.mark.parametrize(
        "content, exception",
        (
            ({"kookdff": "?A?A??"}, exceptions.BadSchemaError),
            ({"test_name": "name", "stages": [{"name": "lflfl"}]}, TypeError),
        ),
    )
    def test_reraise_exception(
        self, tavern_test_content, content: dict, exception: BaseException
    ):
        """Verify that exceptions are properly reraised when loading YAML test files.

        Test that when an exception occurs during test generation, it is properly
        reraised as a schema error if the schema is bad."""

        def raise_error(test_spec):
            raise TypeError

        tavern_test_content.insert(0, content)

        with tavern_test_file(tavern_test_content) as filename:
            with pytest.raises(exception):
                list(
                    YamlFile.collect(
                        Opener(
                            path=filename,
                            _generate_items=raise_error,
                        )
                    )
                )



import contextlib
import logging
from collections.abc import Mapping
from typing import Any, Optional, Protocol

from requests.status_codes import _codes  # type:ignore

from tavern._core import exceptions
from tavern._core.dict_util import deep_dict_merge
from tavern._core.pytest.config import TestConfig
from tavern.response import BaseResponse

logger: logging.Logger = logging.getLogger(__name__)


class ResponseLike(Protocol):
    """Protocol for response-like objects"""

    headers: Mapping[str, str]

    @property
    def text(self) -> str: ...

    def json(self) -> Any: ...


class CommonResponse(BaseResponse):
    """Common response verification functionality shared by REST and GraphQL"""

    def __init__(
        self,
        session,
        name: str,
        expected: dict[str, Any],
        test_block_config: TestConfig,
        default_status_code: int = 200,
        multiple_responses_block: str | None = None,
    ) -> None:
        defaults = {"status_code": default_status_code}
        super().__init__(
            name,
            deep_dict_merge(defaults, expected),
            test_block_config,
            multiple_responses_block=multiple_responses_block,
        )

        def check_code(code: int) -> None:
            if int(code) not in _codes:
                logger.warning("Unexpected status code '%s'", code)

        in_file = self.expected["status_code"]
        try:
            if isinstance(in_file, list):
                for code_ in in_file:
                    check_code(code_)
            else:
                check_code(in_file)
        except TypeError as e:
            raise exceptions.BadSchemaError("Invalid code") from e

    def __str__(self) -> str:
        if self.response:
            return self.response.text.strip()
        else:
            return "<Not run yet>"

    def _verbose_log_response(self, response: ResponseLike) -> None:
        """Verbosely log the response object, with query params etc."""

        logger.info("Response: '%s'", response)

        def log_dict_block(block, name):
            if block:
                to_log = name + ":"

                if isinstance(block, list):
                    for v in block:
                        to_log += f"\n  - {v}"
                elif isinstance(block, dict):
                    for k, v in block.items():
                        to_log += f"\n  {k}: {v}"
                else:
                    to_log += f"\n {block}"
                logger.debug(to_log)

        if hasattr(response, "headers"):
            log_dict_block(response.headers, "Headers")

        with contextlib.suppress(ValueError):
            log_dict_block(response.json(), "Body")

    def _validate_block(
        self, blockname: str, block: Mapping, read_from: Optional[dict] = None
    ) -> None:
        """Validate a block of the response

        Args:
            blockname: which part of the response is being checked
            block: The actual part being checked
            read_from: The block to read from, or self.expected if not specified
        """
        if read_from is None:
            read_from = self.expected

        try:
            expected_block = read_from[blockname]
        except KeyError:
            expected_block = None

        if isinstance(expected_block, dict):
            if expected_block.pop("$ext", None):
                raise exceptions.MisplacedExtBlockException(
                    blockname,
                )

        if blockname == "headers" and expected_block is not None:
            # Special case for headers. These need to be checked in a case
            # insensitive manner
            block = {i.lower(): j for i, j in block.items()}
            expected_block = {i.lower(): j for i, j in expected_block.items()}

        logger.debug("Validating response '%s' against %s", blockname, expected_block)

        test_strictness = self.test_block_config.strict
        if blockname == "data":
            block_strictness = test_strictness.option_for("json")
        else:
            block_strictness = test_strictness.option_for(blockname)

        self.recurse_check_key_match(expected_block, block, blockname, block_strictness)

    def _validate_text(self, response_text: str) -> None:
        """Validate response body as plain text

        Args:
            response_text: The actual response text (response.text)
        """
        expected_text = self.expected.get("text")
        if expected_text is None:
            return

        logger.debug("Validating response text against expected text")
        test_strictness = self.test_block_config.strict
        block_strictness = test_strictness.option_for("text")

        self.recurse_check_key_match(
            expected_text, response_text, "text", block_strictness
        )

    def _common_verify_save(
        self,
        body: Any,
        response: ResponseLike,
    ) -> dict:
        """Common save functionality"""
        saved: dict = {}

        logger.debug(f"Saving response to variables with {body} and {self.expected}")
        if body is not None:
            saved.update(self.maybe_get_save_values_from_save_block("json", body))

        saved.update(self.maybe_get_save_values_from_ext(response, self.expected))

        return saved



import ast

import pytest

from tavern._core import exceptions
from tavern._core.pytest.file import _ast_node_to_literal, _format_test_marks


@pytest.mark.parametrize(
    "marks,expected",
    [
        (["skip"], ([pytest.mark.skip], [])),
        (["xfail"], ([pytest.mark.xfail], [])),
        (["slow"], ([pytest.mark.slow], [])),
        (["skip", "xfail"], ([pytest.mark.skip, pytest.mark.xfail], [])),
        (["xdist_group('group1')"], ([pytest.mark.xdist_group("group1")], [])),
        (
            ["xdist_group(group='group1')"],
            ([pytest.mark.xdist_group(group="group1")], []),
        ),
        (
            ["xdist_group('group1', 'group2')"],
            ([pytest.mark.xdist_group("group1", "group2")], []),
        ),
        (
            ["xdist_group(('group1', 'group2'),)"],
            (
                [
                    pytest.mark.xdist_group(
                        ("group1", "group2"),
                    )
                ],
                [],
            ),
        ),
        (
            ["skip", "xfail", 'xdist_group("group1")'],
            (
                [
                    pytest.mark.skip,
                    pytest.mark.xfail,
                    pytest.mark.xdist_group("group1"),
                ],
                [],
            ),
        ),
        # Additional test cases
        (["usefixtures('fixture1')"], ([pytest.mark.usefixtures("fixture1")], [])),
        (
            ["usefixtures('fixture1', 'fixture2')"],
            ([pytest.mark.usefixtures("fixture1", "fixture2")], []),
        ),
        (
            ["skipif(True, reason='test')"],
            ([pytest.mark.skipif(True, reason="test")], []),
        ),
        (
            ["xfail(reason='flaky')"],
            ([pytest.mark.xfail(reason="flaky")], []),
        ),
        (
            ["parametrize('a,b', [(1,2), (3,4)])"],
            ([pytest.mark.parametrize("a,b", [(1, 2), (3, 4)])], []),
        ),
        (
            ["skip", "slow", 'usefixtures("db")'],
            (
                [
                    pytest.mark.skip,
                    pytest.mark.slow,
                    pytest.mark.usefixtures("db"),
                ],
                [],
            ),
        ),
    ],
)
def test_format_test_marks(marks, expected):
    # Dummy values for fmt_vars and test_name, as required by the function signature
    result = _format_test_marks(marks, fmt_vars={}, test_name="dummy")
    assert result == expected


@pytest.mark.parametrize(
    "invalid_marks",
    [
        ["invalid(mark)"],  # nonexistent mark name
        ["xdist_group('unclosed)"],  # Invalid string literal
        ["xdist_group(missing_quote)"],  # Invalid arg format
    ],
)
def test_failing_format_marks(invalid_marks):
    with pytest.raises(exceptions.BadSchemaError):
        _format_test_marks(invalid_marks, fmt_vars={}, test_name="dummy")


class TestAstNodeToLiteral:
    """Tests for _ast_node_to_literal function"""

    def test_constant_node(self):
        """Test converting ast.Constant nodes"""

        # Test string constant
        node = ast.Constant(value="test_string")
        result = _ast_node_to_literal(node)
        assert result == "test_string"

        # Test numeric constant
        node = ast.Constant(value=42)
        result = _ast_node_to_literal(node)
        assert result == 42

        # Test boolean constant
        node = ast.Constant(value=True)
        result = _ast_node_to_literal(node)
        assert result is True

        # Test None constant
        node = ast.Constant(value=None)
        result = _ast_node_to_literal(node)
        assert result is None

    def test_list_node(self):
        """Test converting ast.List nodes"""

        # Empty list
        node = ast.List(elts=[], ctx=ast.Load())
        result = _ast_node_to_literal(node)
        assert result == []

        # List with constants
        node = ast.List(
            elts=[
                ast.Constant(value="a"),
                ast.Constant(value=1),
                ast.Constant(value=True),
            ],
            ctx=ast.Load(),
        )
        result = _ast_node_to_literal(node)
        assert result == ["a", 1, True]

        # Nested list
        inner_list = ast.List(elts=[ast.Constant(value="nested")], ctx=ast.Load())
        node = ast.List(
            elts=[ast.Constant(value="outer"), inner_list],
            ctx=ast.Load(),
        )
        result = _ast_node_to_literal(node)
        assert result == ["outer", ["nested"]]

    def test_dict_node(self):
        """Test converting ast.Dict nodes"""

        # Empty dict
        node = ast.Dict(keys=[], values=[], ctx=ast.Load())
        result = _ast_node_to_literal(node)
        assert result == {}

        # Dict with constants
        node = ast.Dict(
            keys=[
                ast.Constant(value="key1"),
                ast.Constant(value="key2"),
            ],
            values=[
                ast.Constant(value="value1"),
                ast.Constant(value=2),
            ],
            ctx=ast.Load(),
        )
        result = _ast_node_to_literal(node)
        assert result == {"key1": "value1", "key2": 2}

        # Nested dict
        inner_dict = ast.Dict(
            keys=[ast.Constant(value="inner_key")],
            values=[ast.Constant(value="inner_value")],
            ctx=ast.Load(),
        )
        node = ast.Dict(
            keys=[
                ast.Constant(value="outer_key"),
                ast.Constant(value="nested_dict"),
            ],
            values=[
                ast.Constant(value="outer_value"),
                inner_dict,
            ],
            ctx=ast.Load(),
        )
        result = _ast_node_to_literal(node)
        assert result == {
            "outer_key": "outer_value",
            "nested_dict": {"inner_key": "inner_value"},
        }

    def test_tuple_node(self):
        """Test converting ast.Tuple nodes"""

        # Empty tuple
        node = ast.Tuple(elts=[], ctx=ast.Load())
        result = _ast_node_to_literal(node)
        assert result == ()

        # Tuple with constants
        node = ast.Tuple(
            elts=[
                ast.Constant(value="a"),
                ast.Constant(value=1),
                ast.Constant(value=True),
            ],
            ctx=ast.Load(),
        )
        result = _ast_node_to_literal(node)
        assert result == ("a", 1, True)

        # Nested tuple
        inner_tuple = ast.Tuple(elts=[ast.Constant(value="nested")], ctx=ast.Load())
        node = ast.Tuple(
            elts=[ast.Constant(value="outer"), inner_tuple],
            ctx=ast.Load(),
        )
        result = _ast_node_to_literal(node)
        assert result == ("outer", ("nested",))

    def test_name_node_constants(self):
        """Test converting ast.Name nodes for constants"""

        # Test True
        node = ast.Name(id="True", ctx=ast.Load())
        result = _ast_node_to_literal(node)
        assert result is True

        # Test False
        node = ast.Name(id="False", ctx=ast.Load())
        result = _ast_node_to_literal(node)
        assert result is False

        # Test None
        node = ast.Name(id="None", ctx=ast.Load())
        result = _ast_node_to_literal(node)
        assert result is None

    def test_name_node_variable_reference_error(self):
        """Test that ast.Name nodes for variables raise ValueError"""

        node = ast.Name(id="some_variable", ctx=ast.Load())
        with pytest.raises(
            ValueError, match="Unsupported variable reference: some_variable"
        ):
            _ast_node_to_literal(node)

    def test_unsupported_node_type_error(self):
        """Test that unsupported node types raise ValueError"""

        # Use ast.Expr as an example of an unsupported node type
        node = ast.Expr(value=ast.Constant(value="test"))
        with pytest.raises(
            ValueError, match="Unsupported AST node type: <class 'ast.Expr'>"
        ):
            _ast_node_to_literal(node)


def test_format_test_marks_with_fmt_vars():
    """Test that _format_test_marks correctly substitutes formatted variables in mark arguments"""

    marks = ["skipif('{condition}', reason='{reason}')"]
    fmt_vars = {"condition": "True", "reason": "test_skip"}
    expected = ([pytest.mark.skipif("True", reason="test_skip")], [])

    result = _format_test_marks(marks, fmt_vars=fmt_vars, test_name="dummy")
    assert result == expected



"""Pydantic models for HTTP example responses."""

from pydantic import BaseModel, ConfigDict


class HelloResponse(BaseModel):
    """Response model for /hello/<name> endpoint."""

    data: str


class PingResponse(BaseModel):
    """Response model for /ping endpoint."""

    data: str


class StrictPingResponse(BaseModel):
    """Response model for /ping endpoint that rejects extra fields."""

    model_config = ConfigDict(extra="forbid")
    data: str


class NumberResponse(BaseModel):
    """Response model for /numbers endpoint."""

    number: int


class LoginResponse(BaseModel):
    """Response model for /login endpoint."""

    token: str


class ErrorResponse(BaseModel):
    """Response model for error responses."""

    error: str



import asyncio
import logging
import threading
from typing import Any, Optional

logger = logging.getLogger(__name__)


class ThreadedAsyncLoop(threading.Thread):
    """A context-managed async event loop running in a separate thread."""

    def __init__(self, debug: bool = False):
        """Initialise the threaded async loop.

        Args:
            debug: Whether to enable debug mode for the event loop
        """
        self._loop = asyncio.new_event_loop()
        if debug:
            self._loop.set_debug(True)

        # Thread running the event loop
        super().__init__()

    def run(self):
        """Target function for the thread that runs the event loop."""
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()
        # Close the loop after shutdown
        if not self._loop.is_closed():
            self._loop.close()

    def __enter__(self):
        """Enter the context manager.

        Returns:
            The ThreadedAsyncLoop instance.
        """
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context manager and shut down the event loop."""
        # Schedule the loop to stop in a thread-safe manner
        self._loop.call_soon_threadsafe(self._loop.stop)

        # Join the thread with a timeout to prevent indefinite blocking
        self.join(timeout=5.0)

        # Check if thread is still alive after timeout
        if self.is_alive():
            logger.warning("ThreadedAsyncLoop thread did not terminate within timeout")

        # Ensure the loop is closed
        if not self._loop.is_closed():
            self._loop.close()

    def run_coroutine(self, coro, timeout: Optional[int | float] = None) -> Any:
        """Run a coroutine in the thread's event loop.

        Args:
            coro: The coroutine to run
            timeout: Optional timeout for the operation

        Returns:
            The result of the coroutine
        """
        future = asyncio.run_coroutine_threadsafe(coro, self._loop)
        return future.result(timeout=timeout)



import logging
from contextlib import ExitStack
from functools import cached_property

import box
import gql.transport.exceptions
from gql import FileVar

from tavern._core import exceptions
from tavern._core.dict_util import deep_dict_merge, format_keys
from tavern._core.files import guess_filespec
from tavern._core.formatted_str import FormattedString
from tavern._core.pytest.config import TestConfig
from tavern.request import BaseRequest

from .client import GraphQLClient, GraphQLResponseLike

logger: logging.Logger = logging.getLogger(__name__)


def get_file_arguments(file_args: dict, test_block_config: TestConfig) -> dict:
    """Get correct arguments for anything that should be passed as a file to
    gql

    Args:
        file_args: dict of files to upload
        test_block_config: config for test

    Returns:
        mapping of 'files' block to pass directly to gql
    """

    # Note: Not actually using the opened files
    gql_file_vars = {}
    with ExitStack() as stack:
        for var_name, file_path_or_long_format in file_args.items():
            file_spec, form_field_name, resolved_file_path = guess_filespec(
                file_path_or_long_format, stack, test_block_config
            )
            if form_field_name is None:
                form_field_name = var_name

            if form_field_name in gql_file_vars:
                raise exceptions.BadSchemaError(
                    f"Cannot upload multiple files with the same name '{form_field_name}'"
                )

            gql_file_var = FileVar(
                f=resolved_file_path,
                filename=form_field_name,
                content_type=file_spec.content_type,
            )
            gql_file_vars[var_name] = gql_file_var

    return gql_file_vars


def _format_graphql_request(rspec: dict, variables: dict) -> dict:
    """Format a GraphQL request spec, excluding the query field from formatting.

    GraphQL queries contain curly braces which are mistakenly interpreted as format
    placeholders by the standard format_keys function. This function formats all
    fields except the query field to preserve the GraphQL syntax.

    Args:
        rspec: Request specification dictionary
        variables: Variables to format with

    Returns:
        Formatted request specification with query field unchanged
    """
    formatted_rspec = {}

    for key, value in rspec.items():
        if key == "query":
            # Skip formatting for GraphQL queries to preserve { } syntax
            formatted_rspec[key] = FormattedString(value)
        else:
            # Format all other fields normally
            formatted_rspec[key] = format_keys(value, variables)

    return formatted_rspec


class GraphQLRequest(BaseRequest):
    """GraphQL request implementation"""

    def __init__(
        self, session: GraphQLClient, rspec: dict, test_block_config: TestConfig
    ):
        self.session = session
        self.rspec = rspec
        self.test_block_config = test_block_config

        # Format request spec with test variables, excluding query from formatting
        self._formatted_rspec = _format_graphql_request(
            rspec, test_block_config.variables
        )

        # Validate required fields
        self._validate_request()

    def _validate_request(self):
        """Validate GraphQL request structure"""
        if "query" not in self._formatted_rspec:
            raise exceptions.MissingKeysError(
                "GraphQL request must contain 'query' field"
            )

        if "url" not in self._formatted_rspec:
            raise exceptions.MissingKeysError(
                "GraphQL request must contain 'url' field"
            )

        if self.is_subscription_query and "operation_name" not in self._formatted_rspec:
            raise exceptions.MissingKeysError(
                "operation_name is required for subscription requests"
            )

    @property
    def request_vars(self) -> box.Box:
        """Variables used in the request"""
        return box.Box(
            {
                "url": self._formatted_rspec["url"],
                "query": self._formatted_rspec["query"],
                "variables": self._formatted_rspec.get("variables", {}),
                "operation_name": self._formatted_rspec.get("operation_name"),
                "headers": self._formatted_rspec.get("headers", {}),
            }
        )

    def run(self):
        """Execute GraphQL request"""
        try:
            url = str(self._formatted_rspec["url"])
            query = str(self._formatted_rspec["query"])
            variables = self._formatted_rspec.get("variables", {}) or {}
            operation_name = self._formatted_rspec.get("operation_name")

            headers = {}
            files = self._formatted_rspec.get("files")
            if files:
                variables.update(get_file_arguments(files, self.test_block_config))
            else:
                headers["Content-Type"] = "application/json"

            logger.debug(f"graphql variables: {variables}")

            headers = deep_dict_merge(headers, self._formatted_rspec.get("headers", {}))

            if self.is_subscription_query:
                self.session.start_subscription(url, query, variables, operation_name)
                fake_resp = GraphQLResponseLike(result=None)
                logger.debug(
                    "Subscription '%s' started, fake 101 response", operation_name
                )
                return fake_resp

            # Execute regular GraphQL query/mutation
            response = self.session.make_request(
                url=url,
                query=query,
                variables=variables,
                operation_name=operation_name,
                headers=headers,
                has_files=files is not None,
            )

            logger.debug("GraphQL response: %s", response.text)
            return response

        except gql.transport.exceptions.TransportQueryError as e:
            logger.debug("graphql error while making request: %s", e)
            return GraphQLResponseLike(result=e)
        except Exception as e:
            logger.exception("Error executing GraphQL request")
            raise exceptions.TavernException(f"GraphQL request failed: {e}") from e

    @cached_property
    def is_subscription_query(self) -> bool:
        """Check if the query is a subscription"""
        query = self._formatted_rspec.get("query", "").strip()
        # Simple check for subscription keyword at the start of the query
        return query.lower().startswith("subscription")



import dataclasses

from tavern._plugins.graphql import tavernhook
from tavern._plugins.graphql.client import GraphQLClient


class TestTavernGraphQLPlugin:
    def test_plugin_properties(self):
        assert tavernhook.session_type == GraphQLClient
        assert tavernhook.request_block_name == "graphql_request"
        assert tavernhook.response_block_name == "graphql_response"

    def test_get_expected_from_request_with_response(self, graphql_test_block_config):
        response_block = {"status_code": 200}
        session = GraphQLClient()

        result = tavernhook.get_expected_from_request(
            response_block, graphql_test_block_config, session
        )

        assert result == {"graphql_responses": [{"status_code": 200}]}

    def test_get_expected_from_request_without_response(
        self, graphql_test_block_config
    ):
        response_block = None
        session = GraphQLClient()

        result = tavernhook.get_expected_from_request(
            response_block, graphql_test_block_config, session
        )

        assert result is None

    def test_get_expected_from_request_with_variables(self, graphql_test_block_config):
        response_block = {"status_code": 200, "data": {"user_id": "{user_id}"}}

        test_block_config = dataclasses.replace(
            graphql_test_block_config, variables={"user_id": "123"}
        )

        session = GraphQLClient()

        result = tavernhook.get_expected_from_request(
            response_block, test_block_config, session
        )

        assert result == {
            "graphql_responses": [{"status_code": 200, "data": {"user_id": "123"}}]
        }



import asyncio
from unittest.mock import AsyncMock, Mock, patch

import gql
import pytest
from gql.transport.aiohttp import AIOHTTPTransport

from tavern._plugins.graphql.client import ClientCacheKey, GraphQLClient


class TestGraphQLClient:
    def test_init_with_kwargs(self):
        kwargs = {
            "headers": {"Authorization": "Bearer token"},
            "timeout": 60,
        }
        client = GraphQLClient(**kwargs)
        assert client.default_headers == {"Authorization": "Bearer token"}
        assert client.timeout == 60

    def test_start_subscription_success(self):
        """Test starting a subscription successfully"""
        client = GraphQLClient()

        # Mock the necessary components
        with (
            patch(
                "tavern._plugins.graphql.client.WebsocketsTransport"
            ) as mock_ws_transport,
            patch("tavern._plugins.graphql.client.gql") as mock_gql,
        ):
            # Set up the mock session and subscription generator
            mock_session = AsyncMock()
            mock_generator = AsyncMock()
            mock_session.__aenter__ = AsyncMock(return_value=mock_session)
            mock_session.__aexit__ = AsyncMock(return_value=None)
            mock_session.subscribe = Mock(return_value=mock_generator)

            mock_gql_obj = Mock()
            mock_gql.return_value = mock_gql_obj

            # Start a subscription with valid parameters
            with client:  # Use context manager to set up the event loop
                client.start_subscription(
                    url="https://example.com/graphql",
                    query="subscription Test { test }",
                    variables={"param": "value"},
                    operation_name="TestSubscription",
                )

            # Verify the subscription was created
            assert "TestSubscription" in client._subscriptions
            assert mock_ws_transport.called

    def test_start_subscription_requires_operation_name(self):
        """Test that starting a subscription requires operation_name"""
        client = GraphQLClient()

        with (
            patch("tavern._plugins.graphql.client.WebsocketsTransport"),
            patch("tavern._plugins.graphql.client.Client"),
            patch("tavern._plugins.graphql.client.gql"),
        ):
            # Should raise ValueError when operation_name is None
            with pytest.raises(ValueError) as exc_info:
                client.start_subscription(
                    "ws://example.com", "subscription Test { test }", {}, None
                )
            assert "operation_name required for subscriptions" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_get_next_message_not_found(self):
        """Test getting next message from non-existent subscription"""
        client = GraphQLClient()

        with pytest.raises(ValueError) as exc_info:
            await client.get_next_message("non_existent")
        assert "Subscription with name 'non_existent' not found" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_get_next_message_success(self):
        """Test getting next message from subscription successfully"""
        client = GraphQLClient()

        # Create a mock async generator for the subscription
        async def mock_async_gen():
            yield {"data": {"test": "value"}}

        client._subscriptions["test_op"] = mock_async_gen()

        # Test getting the next message
        with client:  # Use context manager to set up the event loop
            message = await client.get_next_message("test_op")
        assert message == {"data": {"test": "value"}}

    @pytest.mark.asyncio
    async def test_get_next_message_timeout(self):
        """Test getting next message times out"""
        client = GraphQLClient()

        # Create a mock async generator that will timeout
        async def slow_async_gen():
            await asyncio.sleep(10)  # This will cause timeout
            yield {"data": {"test": "value"}}

        # Add the mock generator to subscriptions
        client._subscriptions["slow_op"] = slow_async_gen()

        # Test that timeout is raised
        with client:  # Use context manager to set up the event loop
            with pytest.raises(TimeoutError):
                await client.get_next_message("slow_op", timeout=0.1)

    @pytest.mark.asyncio
    async def test_get_next_message_exception(self):
        """Test getting next message when an exception occurs"""
        client = GraphQLClient()

        # Create a mock async generator that raises an exception
        async def error_async_gen():
            raise Exception("Test error")
            yield

        client._subscriptions["error_op"] = error_async_gen()

        # Test that the exception is propagated
        with client:  # Use context manager to set up the event loop
            with pytest.raises(Exception) as exc_info:
                await client.get_next_message("error_op")
        assert "Test error" in str(exc_info.value)


class TestClientCacheKey:
    """Tests for the ClientCacheKey class"""

    def test_client_cache_key_creation(self):
        """Test creating a ClientCacheKey from components"""
        headers = {"Authorization": "Bearer token", "Content-Type": "application/json"}
        key = ClientCacheKey(
            url="https://example.com/graphql",
            headers=headers,
            timeout=30,
        )
        assert key.url == "https://example.com/graphql"
        assert key.timeout == 30
        # Headers should be converted to sorted tuple
        assert isinstance(key.headers, tuple)
        assert ("Authorization", "Bearer token") in key.headers
        assert ("Content-Type", "application/json") in key.headers

    def test_client_cache_key_hashable(self):
        """Test that ClientCacheKey instances are hashable"""
        headers1 = {"Authorization": "Bearer token"}
        headers2 = {"Authorization": "Bearer token"}
        headers3 = {"Authorization": "Different token"}

        key1 = ClientCacheKey(
            url="https://example.com/graphql", headers=headers1, timeout=30
        )
        key2 = ClientCacheKey(
            url="https://example.com/graphql", headers=headers2, timeout=30
        )
        key3 = ClientCacheKey(
            url="https://example.com/graphql", headers=headers3, timeout=30
        )

        # Keys with same values should have same hash
        assert hash(key1) == hash(key2)

        # Keys with different values should have different hashes
        assert hash(key1) != hash(key3)

    def test_client_cache_key_equality(self):
        """Test ClientCacheKey equality comparison"""
        headers1 = {"Authorization": "Bearer token"}
        headers2 = {"Authorization": "Bearer token"}
        headers3 = {"Authorization": "Different token"}

        key1 = ClientCacheKey(
            url="https://example.com/graphql", headers=headers1, timeout=30
        )
        key2 = ClientCacheKey(
            url="https://example.com/graphql", headers=headers2, timeout=30
        )
        key3 = ClientCacheKey(
            url="https://example.com/graphql", headers=headers3, timeout=30
        )

        # Equal keys
        assert key1 == key2

        # Different keys
        assert key1 != key3
        assert key1 != "not a client cache key"

    def test_client_cache_key_headers_order_independence(self):
        """Test that header order doesn't matter for ClientCacheKey"""
        headers1 = {"Authorization": "Bearer token", "Content-Type": "application/json"}
        headers2 = {"Content-Type": "application/json", "Authorization": "Bearer token"}

        key1 = ClientCacheKey(
            url="https://example.com/graphql", headers=headers1, timeout=30
        )
        key2 = ClientCacheKey(
            url="https://example.com/graphql", headers=headers2, timeout=30
        )

        # Should be equal despite different ordering
        assert key1 == key2
        assert hash(key1) == hash(key2)

    def test_client_cache_key_repr(self):
        """Test ClientCacheKey string representation"""
        headers = {"Authorization": "Bearer token"}
        key = ClientCacheKey(
            url="https://example.com/graphql", headers=headers, timeout=30
        )
        repr_str = repr(key)
        assert "ClientCacheKey" in repr_str
        assert "https://example.com/graphql" in repr_str


class TestClientCaching:
    """Tests for GraphQL client caching functionality"""

    def test_make_request_creates_new_client(self):
        """Test that first request creates a new GraphQL client"""
        client = GraphQLClient()

        query = """
            query TestQuery {
                user {
                    id
                }
            }
        """

        with patch("tavern._plugins.graphql.client.AIOHTTPTransport") as mock_transport:
            mock_transport_instance = Mock(spec=AIOHTTPTransport)
            mock_transport.return_value = mock_transport_instance

            with patch("tavern._plugins.graphql.client.Client") as mock_client:
                mock_execution_result = Mock()
                mock_execution_result.data = {"user": {"id": "1"}}
                mock_client_instance = Mock(spec=gql.Client)
                mock_client_instance.execute.return_value = mock_execution_result
                mock_client.return_value = mock_client_instance

                # Make a request
                response = client.make_request(
                    url="https://example.com/graphql",
                    query=query,
                )

                # Verify client and transport were created
                assert mock_transport.called
                assert mock_client.called
                assert len(client._gql_client_cache) == 1
                assert len(client._transport_cache) == 1

                # Verify the response
                assert response.json() == {"user": {"id": "1"}}

    def test_make_request_reuses_cached_client(self):
        """Test that subsequent requests to same URL reuse GraphQL client"""
        client = GraphQLClient()

        query1 = """
            query Query1 {
                user {
                    id
                }
            }
        """

        query2 = """
            query Query2 {
                posts {
                    id
                }
            }
        """

        with patch("tavern._plugins.graphql.client.AIOHTTPTransport") as mock_transport:
            mock_transport_instance = Mock(spec=AIOHTTPTransport)
            mock_transport.return_value = mock_transport_instance

            with patch("tavern._plugins.graphql.client.Client") as mock_client:
                mock_execution_result = Mock()
                mock_execution_result.data = {"data": "result"}
                mock_client_instance = Mock(spec=gql.Client)
                mock_client_instance.execute.return_value = mock_execution_result
                mock_client.return_value = mock_client_instance

                # Make first request
                client.make_request(
                    url="https://example.com/graphql",
                    query=query1,
                )

                # Make second request to same URL
                client.make_request(
                    url="https://example.com/graphql",
                    query=query2,
                )

                # Verify client was created only once
                assert mock_transport.call_count == 1
                assert mock_client.call_count == 1
                assert len(client._gql_client_cache) == 1
                assert len(client._transport_cache) == 1

    def test_make_request_different_urls_create_different_clients(self):
        """Test that requests to different URLs create different GraphQL clients"""
        client = GraphQLClient()

        query = """
            query TestQuery {
                user {
                    id
                }
            }
        """

        with patch("tavern._plugins.graphql.client.AIOHTTPTransport") as mock_transport:
            mock_transport_instance = Mock(spec=AIOHTTPTransport)
            mock_transport.return_value = mock_transport_instance

            with patch("tavern._plugins.graphql.client.Client") as mock_client:
                mock_execution_result = Mock()
                mock_execution_result.data = {"data": "result"}
                mock_client_instance = Mock(spec=gql.Client)
                mock_client_instance.execute.return_value = mock_execution_result
                mock_client.return_value = mock_client_instance

                # Make requests to different URLs
                client.make_request(
                    url="https://example.com/graphql",
                    query=query,
                )

                client.make_request(
                    url="https://another-example.com/graphql",
                    query=query,
                )

                # Verify two clients were created
                assert mock_transport.call_count == 2
                assert mock_client.call_count == 2
                assert len(client._gql_client_cache) == 2
                assert len(client._transport_cache) == 2

    def test_make_request_different_headers_create_different_clients(self):
        """Test that requests with different headers create different GraphQL clients"""
        client = GraphQLClient()

        query = """
            query TestQuery {
                user {
                    id
                }
            }
        """

        with patch("tavern._plugins.graphql.client.AIOHTTPTransport") as mock_transport:
            mock_transport_instance = Mock(spec=AIOHTTPTransport)
            mock_transport.return_value = mock_transport_instance

            with patch("tavern._plugins.graphql.client.Client") as mock_client:
                mock_execution_result = Mock()
                mock_execution_result.data = {"data": "result"}
                mock_client_instance = Mock(spec=gql.Client)
                mock_client_instance.execute.return_value = mock_execution_result
                mock_client.return_value = mock_client_instance

                # Make first request with Authorization header
                client.make_request(
                    url="https://example.com/graphql",
                    query=query,
                    headers={"Authorization": "Bearer token1"},
                )

                # Make second request with different Authorization header
                client.make_request(
                    url="https://example.com/graphql",
                    query=query,
                    headers={"Authorization": "Bearer token2"},
                )

                # Verify two clients were created (different headers)
                assert mock_transport.call_count == 2
                assert mock_client.call_count == 2
                assert len(client._gql_client_cache) == 2
                assert len(client._transport_cache) == 2

    def test_make_request_same_headers_reuse_client(self):
        """Test that requests with same headers reuse GraphQL client"""
        client = GraphQLClient()

        query = """
            query TestQuery {
                user {
                    id
                }
            }
        """

        with patch("tavern._plugins.graphql.client.AIOHTTPTransport") as mock_transport:
            mock_transport_instance = Mock(spec=AIOHTTPTransport)
            mock_transport.return_value = mock_transport_instance

            with patch("tavern._plugins.graphql.client.Client") as mock_client:
                mock_execution_result = Mock()
                mock_execution_result.data = {"data": "result"}
                mock_client_instance = Mock(spec=gql.Client)
                mock_client_instance.execute.return_value = mock_execution_result
                mock_client.return_value = mock_client_instance

                # Make two requests with same headers
                client.make_request(
                    url="https://example.com/graphql",
                    query=query,
                    headers={"Authorization": "Bearer token"},
                )

                client.make_request(
                    url="https://example.com/graphql",
                    query=query,
                    headers={"Authorization": "Bearer token"},
                )

                # Verify only one client was created
                assert mock_transport.call_count == 1
                assert mock_client.call_count == 1
                assert len(client._gql_client_cache) == 1
                assert len(client._transport_cache) == 1

    def test_client_context_manager_closes_cached_clients(self):
        """Test that context manager exit closes cached clients and transports"""
        client = GraphQLClient()

        with patch("tavern._plugins.graphql.client.AIOHTTPTransport") as mock_transport:
            mock_transport_instance = Mock(spec=AIOHTTPTransport)
            mock_transport.return_value = mock_transport_instance

            with patch("tavern._plugins.graphql.client.Client") as mock_client:
                mock_execution_result = Mock()
                mock_execution_result.data = {"data": "result"}
                mock_client_instance = Mock(spec=gql.Client)
                mock_client_instance.execute.return_value = mock_execution_result
                mock_client.return_value = mock_client_instance

                # Use context manager and make a request
                with client:
                    client.make_request(
                        url="https://example.com/graphql",
                        query="query { test }",
                    )

                # Verify transport.close() was called on exit
                assert mock_transport_instance.close.called
                # Verify client cache was populated
                assert len(client._gql_client_cache) == 1
                assert len(client._transport_cache) == 1
